import java.util.*;

public class CyclicBarrier {

    private int arrived = 0;
    private int N;

    //
    // Integer is good enough to keep track of active turns.
    // As of today, there is not even possible to have that
    // many active threads in Java.
    // The use of HashMap is perhaps an overkill.  Usually,
    // there are only a few active turns needed and other more
    // efficient implementations could be perhaps possible.
    //
    private Map<Integer,Integer> flag = new HashMap<Integer,Integer>();
    private int turn;
    
    public CyclicBarrier(int N) {
	this.N = N;
	turn = 0;
	flag.put(turn, 0);
    }
  
    public synchronized void await() throws InterruptedException {
	arrived++;
	if (arrived < N) {
	    int myTurn = turn;
	    while (flag.get(myTurn) == 0)
		wait();
	    if (flag.put(myTurn, flag.get(myTurn)-1) == 1)
		flag.remove(myTurn);
	}
	else {
	    flag.put(turn, N-1);
	    turn++;
	    flag.put(turn, 0);
	    notifyAll();
	    arrived = 0;
	}
    }
}
